/*
 * Copyright (c) 2007-2008 Intel Corporation.  All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: $
 */
#include <stdio.h>
#include <stdlib.h>

#if defined(_WIN32) || defined(_WIN64)
#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <io.h>
#include <process.h>
#include <complib/cl_types.h>
#define __BYTE_ORDER __LITTLE_ENDIAN

#define getpid _getpid
#define F64x "%I64x"
#define DAPL_PROVIDER "ibnic0v2"
#else
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <inttypes.h>
#include <string.h>
#include <unistd.h>

#define DAPL_PROVIDER "ofa-v2-ib0"
#define F64x "%"PRIx64""

#endif

#include "dat2/udat.h"
#include "dat2/dat_ib_extensions.h"

int disconnect_ep(void);

#define _OK(status, str) \
{\
	const char  *maj_msg, *min_msg;\
	if (status != DAT_SUCCESS) {\
		dat_strerror(status, &maj_msg, &min_msg);\
		fprintf(stderr, str " returned %s : %s\n", maj_msg, min_msg);\
		exit(1);\
	}\
}

#define _OK2(status, str)\
{\
	const char  *maj_msg, *min_msg;\
	if (status != DAT_SUCCESS) {\
		dat_strerror(status, &maj_msg, &min_msg);\
		fprintf(stderr, str " returned %s : %s\n", maj_msg, min_msg);\
		exit(1);\
	}\
}

/* byte swap helpers from Complib */
#if __BYTE_ORDER == __LITTLE_ENDIAN
#define ntoh16(x) (uint16_t)( \
        (((uint16_t)(x) & 0x00FF) << 8) | \
        (((uint16_t)(x) & 0xFF00) >> 8))
#define hton16(x) ntoh16(x)
#define ntoh32(x) (uint32_t)( \
        (((uint32_t)(x) & 0x000000FF) << 24)| \
        (((uint32_t)(x) & 0x0000FF00) << 8) | \
        (((uint32_t)(x) & 0x00FF0000) >> 8) | \
        (((uint32_t)(x) & 0xFF000000) >> 24))
#define hton32(x) ntoh32(x)
#define ntoh64(x) (uint64_t)( \
        (((uint64_t)x & 0x00000000000000FF) << 56) | \
        (((uint64_t)x & 0x000000000000FF00) << 40) | \
        (((uint64_t)x & 0x0000000000FF0000) << 24) | \
        (((uint64_t)x & 0x00000000FF000000) << 8 ) | \
        (((uint64_t)x & 0x000000FF00000000) >> 8 ) | \
        (((uint64_t)x & 0x0000FF0000000000) >> 24) | \
        (((uint64_t)x & 0x00FF000000000000) >> 40) | \
        (((uint64_t)x & 0xFF00000000000000) >> 56))
#define hton64(x) ntoh64(x)
#elif __BYTE_ORDER == __BIG_ENDIAN
#define hton16(x) (x)
#define ntoh16(x) (x)
#define hton32(x) (x)
#define ntoh32(x) (x)
#define hton64(x) (x)
#define ntoh64(x) (x)
#endif  /* __BYTE_ORDER == __BIG_ENDIAN */

#define DTO_TIMEOUT       (1000*1000*5)
#define CONN_TIMEOUT      (1000*1000*10)
#define SERVER_TIMEOUT    (1000*1000*120)
#define SERVER_CONN_QUAL	31111
#define BUF_SIZE		256
#define BUF_SIZE_ATOMIC		8
#define REG_MEM_COUNT		10
#define SND_RDMA_BUF_INDEX	0
#define RCV_RDMA_BUF_INDEX	1
#define SEND_BUF_INDEX		2
#define RECV_BUF_INDEX		3

DAT_VADDR 		*atomic_buf;
DAT_LMR_HANDLE		lmr_atomic;
DAT_LMR_CONTEXT		lmr_atomic_context;
DAT_RMR_CONTEXT		rmr_atomic_context;
DAT_VLEN		reg_atomic_size;
DAT_VADDR		reg_atomic_addr;
DAT_LMR_HANDLE		lmr[ REG_MEM_COUNT ];
DAT_LMR_CONTEXT		lmr_context[ REG_MEM_COUNT ];
DAT_RMR_TRIPLET		rmr[ REG_MEM_COUNT ];
DAT_RMR_CONTEXT		rmr_context[ REG_MEM_COUNT ];
DAT_VLEN		reg_size[ REG_MEM_COUNT ];
DAT_VADDR		reg_addr[ REG_MEM_COUNT ];
DAT_RMR_TRIPLET *	buf[ REG_MEM_COUNT ];
DAT_EP_HANDLE		ep;
DAT_EVD_HANDLE		async_evd = DAT_HANDLE_NULL;
DAT_IA_HANDLE		ia = DAT_HANDLE_NULL;
DAT_PZ_HANDLE		pz = DAT_HANDLE_NULL;
DAT_EVD_HANDLE		cr_evd = DAT_HANDLE_NULL;
DAT_EVD_HANDLE		con_evd = DAT_HANDLE_NULL;
DAT_EVD_HANDLE		dto_evd = DAT_HANDLE_NULL;
DAT_PSP_HANDLE		psp = DAT_HANDLE_NULL;
DAT_CR_HANDLE		cr = DAT_HANDLE_NULL;
int			server;

char *usage = "-s | hostname (default == -s)\n";

#if defined(_WIN32) || defined(_WIN64)
static void sleep(int secs)
{
	Sleep(secs*1000);
}
#define _WSACleanup() WSACleanup()
#else
#define	_WSACleanup()
#endif

void
send_msg(
	void *data,
	DAT_COUNT size,
	DAT_LMR_CONTEXT context,
	DAT_DTO_COOKIE cookie,
	DAT_COMPLETION_FLAGS flags)
{
	DAT_LMR_TRIPLET			iov;
	DAT_EVENT			event;
	DAT_COUNT			nmore;
	DAT_RETURN			status;
	DAT_DTO_COMPLETION_EVENT_DATA *dto_event = 
		&event.event_data.dto_completion_event_data;

	iov.lmr_context     = context;
	iov.virtual_address = (DAT_VADDR)data;
	iov.segment_length  = (DAT_VLEN)size;

	status = dat_ep_post_send(ep,
				   1,
				   &iov,
				   cookie,
				   flags);
	_OK(status, "dat_ep_post_send");

	if (! (flags & DAT_COMPLETION_SUPPRESS_FLAG)) {
		status = dat_evd_wait(dto_evd, DTO_TIMEOUT, 
				      1, &event, &nmore);
		_OK(status, "dat_evd_wait after dat_ep_post_send");

		if (event.event_number != DAT_DTO_COMPLETION_EVENT) {
			printf("unexpected event waiting for post_send "
				"completion - 0x%x\n", event.event_number);
			exit(1);
		}

		_OK(dto_event->status, "event status for post_send");
	}
}

int
connect_ep(char *hostname)
{
	DAT_SOCK_ADDR		remote_addr;
	DAT_EP_ATTR		ep_attr;
	DAT_RETURN		status;
	DAT_REGION_DESCRIPTION	region;
	DAT_EVENT		event;
	DAT_COUNT		nmore;
	DAT_LMR_TRIPLET		iov;
	DAT_RMR_TRIPLET		*r_iov;
	DAT_DTO_COOKIE		cookie;
	int			i;
        DAT_CR_ARRIVAL_EVENT_DATA *cr_event = 
		&event.event_data.cr_arrival_event_data;
	DAT_DTO_COMPLETION_EVENT_DATA *dto_event = 
		&event.event_data.dto_completion_event_data;

	status = dat_ia_open(DAPL_PROVIDER, 8, &async_evd, &ia);
	_OK(status, "dat_ia_open");

	status = dat_pz_create(ia, &pz);
	_OK(status, "dat_pz_create");

	status = dat_evd_create(ia, 10, DAT_HANDLE_NULL, DAT_EVD_CR_FLAG,
				&cr_evd );
	_OK(status, "dat_evd_create CR");
	status = dat_evd_create(ia, 10, DAT_HANDLE_NULL,
				DAT_EVD_CONNECTION_FLAG, &con_evd );
	_OK(status, "dat_evd_create CR");
	status = dat_evd_create(ia, 10, DAT_HANDLE_NULL, DAT_EVD_DTO_FLAG,
				&dto_evd );
	_OK(status, "dat_evd_create DTO");

	memset(&ep_attr, 0, sizeof(ep_attr));
	ep_attr.service_type                = DAT_SERVICE_TYPE_RC;
	ep_attr.max_rdma_size               = 0x10000;
	ep_attr.qos                         = 0;
	ep_attr.recv_completion_flags       = 0;
	ep_attr.max_recv_dtos               = 10;
	ep_attr.max_request_dtos            = 10;
	ep_attr.max_recv_iov                = 1;
	ep_attr.max_request_iov             = 1;
	ep_attr.max_rdma_read_in            = 4;
	ep_attr.max_rdma_read_out           = 4;
	ep_attr.request_completion_flags    = DAT_COMPLETION_DEFAULT_FLAG;
	ep_attr.ep_transport_specific_count = 0;
	ep_attr.ep_transport_specific       = NULL;
	ep_attr.ep_provider_specific_count  = 0;
	ep_attr.ep_provider_specific        = NULL;

	status = dat_ep_create(ia, pz, dto_evd, dto_evd, 
			       con_evd, &ep_attr, &ep);
	_OK(status, "dat_ep_create");

	for (i = 0; i < REG_MEM_COUNT; i++) {
		buf[ i ] = (DAT_RMR_TRIPLET*)malloc(BUF_SIZE);
		region.for_va = buf[ i ];
		status = dat_lmr_create(ia,
					DAT_MEM_TYPE_VIRTUAL,
					region,
					BUF_SIZE,
					pz,
					DAT_MEM_PRIV_ALL_FLAG|
						DAT_IB_MEM_PRIV_REMOTE_ATOMIC,
					DAT_VA_TYPE_VA,
					&lmr[ i ],
					&lmr_context[ i ],
					&rmr_context[ i ],
					&reg_size[ i ],
					&reg_addr[ i ]);
		_OK(status, "dat_lmr_create");
	}

	/* register atomic return buffer for original data */
	atomic_buf = (DAT_UINT64*)malloc(BUF_SIZE);
	region.for_va = atomic_buf;
	status = dat_lmr_create(ia,
				DAT_MEM_TYPE_VIRTUAL,
				region,
				BUF_SIZE_ATOMIC,
				pz,
				DAT_MEM_PRIV_ALL_FLAG|
					DAT_IB_MEM_PRIV_REMOTE_ATOMIC,
				DAT_VA_TYPE_VA,
				&lmr_atomic,
				&lmr_atomic_context,
				&rmr_atomic_context,
				&reg_atomic_size,
				&reg_atomic_addr);
	_OK(status, "dat_lmr_create atomic");
	
	for (i = RECV_BUF_INDEX; i < REG_MEM_COUNT; i++) {
		cookie.as_64        = i;
		iov.lmr_context     = lmr_context[ i ];
		iov.virtual_address = (DAT_VADDR) buf[ i ];
		iov.segment_length  = BUF_SIZE;

		status = dat_ep_post_recv(ep,
					   1,
					   &iov,
					   cookie,
					   DAT_COMPLETION_DEFAULT_FLAG);
		_OK(status, "dat_ep_post_recv");
	}

	/* setup receive buffer to initial string to be overwritten */
	strcpy((char*)buf[ RCV_RDMA_BUF_INDEX ], "blah, blah, blah\n");

	if (server) {
		strcpy((char*)buf[ SND_RDMA_BUF_INDEX ],"server written data");
		status = dat_psp_create(ia,
					SERVER_CONN_QUAL,
					 cr_evd,
					 DAT_PSP_CONSUMER_FLAG,
					 &psp);
		_OK(status, "dat_psp_create");

		printf("Server waiting for connect request\n");
		status = dat_evd_wait(cr_evd, SERVER_TIMEOUT, 
				      1, &event, &nmore);
		_OK(status, "listen dat_evd_wait");

		if (event.event_number != DAT_CONNECTION_REQUEST_EVENT) {
			printf("unexpected event after dat_psp_create: 0x%x\n",
				event.event_number); 
			exit(1);
		}

		if ((cr_event->conn_qual != SERVER_CONN_QUAL) ||
			(cr_event->sp_handle.psp_handle != psp)) {
			printf("wrong cr event data\n");
			exit(1);
		}

		cr = cr_event->cr_handle;
		status = dat_cr_accept(cr, ep, 0, (DAT_PVOID)0);
		printf("Server waiting for accept response\n");

	} else {
		struct addrinfo	*target;

		if (getaddrinfo (hostname, NULL, NULL, &target) != 0) {
			printf("Error getting remote address.\n");
			exit(1);
		}

		printf ("Server Name: %s \n", hostname);
		printf ("Server Net Address: %s\n", inet_ntoa(
			((struct sockaddr_in *)target->ai_addr)->sin_addr));

		remote_addr = *((DAT_IA_ADDRESS_PTR)target->ai_addr);
		freeaddrinfo(target);
		strcpy((char*)buf[ SND_RDMA_BUF_INDEX ],"client written data");
		status = dat_ep_connect(ep,
					 &remote_addr,
					 SERVER_CONN_QUAL,
					 CONN_TIMEOUT,
					 0,
					 (DAT_PVOID)0,
					 0,
					 DAT_CONNECT_DEFAULT_FLAG );
		_OK(status, "dat_psp_create");
		printf("Client waiting for connect response\n");
	}

	status = dat_evd_wait(con_evd, CONN_TIMEOUT, 1, &event, &nmore);
		_OK(status, "connect dat_evd_wait");

	if (event.event_number != DAT_CONNECTION_EVENT_ESTABLISHED) {
		printf("unexpected event after dat_ep_connect: 0x%x\n", 
			event.event_number); 
		exit(1);
	}

	printf("Connected!\n");

	/*
	 *  Setup our remote memory and tell the other side about it
         *  Swap to network order.
	 */
        r_iov = (DAT_RMR_TRIPLET*)buf[SEND_BUF_INDEX];
	r_iov->rmr_context = hton32(rmr_context[RCV_RDMA_BUF_INDEX]);
	r_iov->virtual_address = hton64((DAT_VADDR)buf[RCV_RDMA_BUF_INDEX]);
	r_iov->segment_length = hton32(BUF_SIZE);

        printf("%d Send RMR msg to remote: r_key_ctx=0x%x,va=%p,len=0x%x\n",
               getpid(), hton32(r_iov->rmr_context), 
               hton64(r_iov->virtual_address), hton32(r_iov->segment_length)); 

	send_msg(buf[SEND_BUF_INDEX],
		 sizeof(DAT_RMR_TRIPLET),
		 lmr_context[SEND_BUF_INDEX],
		 cookie,
		 DAT_COMPLETION_SUPPRESS_FLAG);

	/*
	 *  Wait for their RMR
	 */
	printf("Waiting for remote to send RMR data\n");
	status = dat_evd_wait(dto_evd, DTO_TIMEOUT, 1, &event, &nmore);
	_OK(status, "dat_evd_wait after dat_ep_post_send");

	if (event.event_number != DAT_DTO_COMPLETION_EVENT) {
		printf("unexpected event waiting for RMR context - 0x%x\n", 
			event.event_number);
		exit(1);
	}

	_OK(dto_event->status, "event status for post_send");
	if ((dto_event->transfered_length != sizeof(DAT_RMR_TRIPLET)) ||
	     (dto_event->user_cookie.as_64 != RECV_BUF_INDEX)) {
		printf("unexpected event data for receive: len=%d cookie=%d "
			"expected %d/%d\n",
			(int)dto_event->transfered_length,
			(int)dto_event->user_cookie.as_64,
			sizeof(DAT_RMR_TRIPLET), RECV_BUF_INDEX);
		exit(1);
	}

	/* swap RMR,address info to host order */
        r_iov = (DAT_RMR_TRIPLET*)buf[RECV_BUF_INDEX];
	r_iov->rmr_context = ntoh32(r_iov->rmr_context);
	r_iov->virtual_address = ntoh64(r_iov->virtual_address);
	r_iov->segment_length =	ntoh32(r_iov->segment_length);

	printf("%d Received RMR from remote: "
		"r_iov: r_key_ctx=%x,va="F64x",len=0x%x\n",
		getpid(), r_iov->rmr_context,
		r_iov->virtual_address,
		r_iov->segment_length);

	return(0);
}

int
disconnect_ep(void)
{
	DAT_RETURN	status;
	int		i;
	DAT_EVENT	event;
	DAT_COUNT	nmore;

	status = dat_ep_disconnect(ep, DAT_CLOSE_DEFAULT);
	_OK2(status, "dat_ep_disconnect");

   	status = dat_evd_wait(con_evd, DAT_TIMEOUT_INFINITE, 1,
					&event, &nmore);
	_OK(status, "dat_evd_wait");
    	
	if (server) {
		status = dat_psp_free(psp);
		_OK2(status, "dat_psp_free");
	}

	for (i = 0; i < REG_MEM_COUNT; i++) {
		status = dat_lmr_free(lmr[ i ]);
		_OK2(status, "dat_lmr_free");
	}

	if (lmr_atomic) {
		status = dat_lmr_free(lmr_atomic);
		_OK2(status, "dat_lmr_free_atomic");
	}

	status = dat_ep_free(ep);
	_OK2(status, "dat_ep_free");

	status = dat_evd_free(dto_evd);
	_OK2(status, "dat_evd_free DTO");
	status = dat_evd_free(con_evd);
	_OK2(status, "dat_evd_free CON");
	status = dat_evd_free(cr_evd);
	_OK2(status, "dat_evd_free CR");

	status = dat_pz_free(pz);
	_OK2(status, "dat_pz_free");

	status = dat_ia_close(ia, DAT_CLOSE_DEFAULT);
	_OK2(status, "dat_ia_close");

	return(0);
}

int
do_immediate()
{
	DAT_REGION_DESCRIPTION	region;
	DAT_EVENT		event;
	DAT_COUNT		nmore;
	DAT_LMR_TRIPLET		iov;
	DAT_RMR_TRIPLET		r_iov;
	DAT_DTO_COOKIE		cookie;
	DAT_RMR_CONTEXT		their_context;
	DAT_RETURN		status;
	DAT_UINT32		immed_data;
	DAT_UINT32		immed_data_recv;
	DAT_DTO_COMPLETION_EVENT_DATA *dto_event = 
		&event.event_data.dto_completion_event_data;
	DAT_IB_EXTENSION_EVENT_DATA *ext_event = 
		(DAT_IB_EXTENSION_EVENT_DATA *)&event.event_extension_data[0];

	printf("\nDoing RDMA WRITE IMMEDIATE DATA\n");

	if (server) {
		immed_data = 0x1111;
	} else {
		immed_data = 0x7777;
	}

	cookie.as_64 = 0x5555;
	
        /* RMR info already swapped back to host order in connect_ep */
	r_iov = *buf[RECV_BUF_INDEX];

	iov.lmr_context     = lmr_context[SND_RDMA_BUF_INDEX];
	iov.virtual_address = (DAT_VADDR) buf[SND_RDMA_BUF_INDEX];
	iov.segment_length  = BUF_SIZE;

	cookie.as_64 = 0x9999;
	
	status = dat_ib_post_rdma_write_immed(ep,		// ep_handle
					      1,		// segments
					      &iov,		// LMR
					      cookie,		// user_cookie
					      &r_iov,		// RMR
					      immed_data,
					      DAT_COMPLETION_DEFAULT_FLAG);
	_OK(status, "dat_ib_post_rdma_write_immed");

	/*
	 *  Collect first event, write completion or inbound recv with immed
	 */
	status = dat_evd_wait(dto_evd, DTO_TIMEOUT, 1, &event, &nmore);
	_OK(status, "dat_evd_wait after dat_ib_post_rdma_write");
	if (event.event_number != DAT_IB_DTO_EVENT)
	{
		printf("unexpected event #0x%x waiting for WR-IMMED #0x%x\n", 
			event.event_number,DAT_IB_DTO_EVENT);
		exit(1);
	}
	
	if (nmore) printf("%s() nmore %d\n",__FUNCTION__,nmore);
	_OK(dto_event->status, "DTO event status");
	if (ext_event->type == DAT_IB_RDMA_WRITE_IMMED)
	{
	  if ((dto_event->transfered_length != BUF_SIZE) ||
	     (dto_event->user_cookie.as_64 != 0x9999)) 
	  {
              printf("unexpected event data for rdma_write_immed: len=%d "
			"cookie=0x%x\n",
			(int)dto_event->transfered_length,
			(int)dto_event->user_cookie.as_64);
	      exit(1);
	  }
	} 
	else if (ext_event->type == DAT_IB_RDMA_WRITE_IMMED_DATA)
        {
	  if ((dto_event->transfered_length != BUF_SIZE) ||
	     (dto_event->user_cookie.as_64 != RECV_BUF_INDEX+1)) 
	  {
		printf("unexpected event data of immediate write: len=%d "
			"cookie=%d expected %d/%d\n",
			(int)dto_event->transfered_length,
			(int)dto_event->user_cookie.as_64,
			sizeof(int), RECV_BUF_INDEX+1);
		exit(1);
	  }
	  
	  /* get immediate data from event */
	  immed_data_recv = ext_event->val.immed.data;
	}
	else
	{
		printf("unexpected extension type for event - 0x%x, 0x%x\n", 
			event.event_number, ext_event->type);
		exit(1);
	}


	/*
	 * Collect second event, write completion or inbound recv with immed
	 */
	status = dat_evd_wait(dto_evd, DTO_TIMEOUT, 1, &event, &nmore);
	_OK(status, "dat_evd_wait after dat_ib_post_rdma_write");
	if (event.event_number != DAT_IB_DTO_EVENT)
	{
		printf("unexpected event # waiting for WR-IMMED - 0x%x\n", 
			event.event_number);
		exit(1);
	}
	
	_OK(dto_event->status, "event status");
	if (ext_event->type == DAT_IB_RDMA_WRITE_IMMED)
	{
	  if ((dto_event->transfered_length != BUF_SIZE) ||
	     (dto_event->user_cookie.as_64 != 0x9999)) 
	  {
              printf("unexpected event data for rdma_write_immed: len=%d "
			"cookie=0x%x\n",
			(int)dto_event->transfered_length,
			(int)dto_event->user_cookie.as_64);
	      exit(1);
	  }
	} 
	else if (ext_event->type == DAT_IB_RDMA_WRITE_IMMED_DATA)
        {
	  if ((dto_event->transfered_length != BUF_SIZE) ||
	     (dto_event->user_cookie.as_64 != RECV_BUF_INDEX+1)) 
	  {
		printf("unexpected event data of immediate write: len=%d "
			"cookie=%d expected %d/%d\n",
			(int)dto_event->transfered_length,
			(int)dto_event->user_cookie.as_64,
			sizeof(int), RECV_BUF_INDEX+1);
		exit(1);
	  }
	  
	  /* get immediate data from event */
	  immed_data_recv = ext_event->val.immed.data;
	}
	else
	{
		printf("unexpected extension type for event - 0x%x, 0x%x\n", 
			event.event_number, ext_event->type);
		exit(1);
	}
		
	if ((server) && (immed_data_recv != 0x7777))
	{
		printf("ERROR: Server: unexpected imm_data_recv 0x%x/0x%x\n", 
			0x7777,  immed_data_recv);
		exit(1);
	} 
	else if ((!server) && (immed_data_recv != 0x1111))
	{
		printf("ERROR: Client: unexpected imm_data_recv 0x%x/0x%x\n", 
			0x1111,  immed_data_recv);
		exit(1);
	}

	if (server)
		printf("Server received immed_data=0x%x\n",immed_data_recv);
	else
		printf("Client received immed_data=0x%x\n",immed_data_recv);
	
	printf("rdma buffer %p contains: %s\n", 
		buf[ RCV_RDMA_BUF_INDEX ], buf[ RCV_RDMA_BUF_INDEX ]);

	printf("\n RDMA_WRITE_WITH_IMMEDIATE_DATA test - PASSED\n");
	return (0);
}

int
do_cmp_swap()
{
	DAT_DTO_COOKIE	cookie;
	DAT_RETURN		status;
	DAT_EVENT		event;
	DAT_COUNT		nmore;
	DAT_LMR_TRIPLET		l_iov;
	DAT_RMR_TRIPLET		r_iov;
	volatile DAT_UINT64	*target = (DAT_UINT64*)buf[RCV_RDMA_BUF_INDEX];
	DAT_DTO_COMPLETION_EVENT_DATA *dto_event = 
		&event.event_data.dto_completion_event_data;
	DAT_IB_EXTENSION_EVENT_DATA *ext_event = 
		(DAT_IB_EXTENSION_EVENT_DATA *)&event.event_extension_data[0];

	printf("\nDoing CMP and SWAP\n");
        
        /* RMR info already swapped back to host order in connect_ep */
	r_iov = *buf[ RECV_BUF_INDEX ];
	
	l_iov.lmr_context     = lmr_atomic_context;
	l_iov.virtual_address = (DAT_UINT64)atomic_buf;
	l_iov.segment_length  = BUF_SIZE_ATOMIC;

	cookie.as_64 = 3333;

	if (server) {
		*target = 0x12345;
		sleep(1);
		/* server does not compare and should not swap */
		printf("dtx svr - starting cmp_swap\n");
		status = dat_ib_post_cmp_and_swap(
						ep, 
						(DAT_UINT64)0x654321, 
						(DAT_UINT64)0x6789A, 
						&l_iov,
						cookie, 
						&r_iov, 
						DAT_COMPLETION_DEFAULT_FLAG);
		printf("dtx svr - done cmp_swap, chk status\n");
	} else {
		*target = 0x54321;
		sleep(1); 
		printf("dtx cli - starting cmp_swap\n");
		/* client does compare and should swap */
		status = dat_ib_post_cmp_and_swap(
						ep, 
						(DAT_UINT64)0x12345, 
						(DAT_UINT64)0x98765,
						&l_iov,
						cookie, 
						&r_iov, 
						DAT_COMPLETION_DEFAULT_FLAG);
		printf("dtx cli - done cmp_swap, chk status\n");
	}
	_OK(status, "dat_ib_post_cmp_and_swap");
	status = dat_evd_wait(dto_evd, DTO_TIMEOUT, 1, &event, &nmore);
	_OK(status, "dat_evd_wait for compare and swap");
	if (event.event_number != DAT_IB_DTO_EVENT) {
	    printf("unexpected event after post_cmp_and_swap: 0x%x\n", 
		    event.event_number); 
	    exit(1);
	}

	_OK(dto_event->status, "event status for CMP and SWAP");
	if (ext_event->type != DAT_IB_CMP_AND_SWAP) {
  	    printf("unexpected event data of cmp_swap: type=%d cookie=%d "
		"original "F64x"\n",
		(int)ext_event->type,
		(int)dto_event->user_cookie.as_64,
		*atomic_buf);
	    exit(1);
	}

	sleep(2); /* wait for other side to complete swap */

	if (server) {
	    printf("Server got original data        = "F64x", expected "
			"0x54321\n", *atomic_buf);
	    printf("Client final result (on server) = "F64x", expected "
			"0x98765\n", *target);

	    if (*atomic_buf != 0x54321 || *target != 0x98765) {
	    	printf("ERROR: Server CMP_SWAP\n");
		exit(1);
	    }
	} else {
	    printf("Client got original data        = "F64x", expected "
			"0x12345\n",*atomic_buf);
	    printf("Server final result (on client) = 0x"F64x", expected "
			"0x54321\n", *target);

	    if (*atomic_buf != 0x12345 || *target != 0x54321) {
	    	printf("ERROR: Client CMP_SWAP\n");
		exit(1);
	    }
	}
	printf("\n CMP_SWAP test - PASSED\n");
	return(0);
}

int
do_fetch_add()
{
	DAT_DTO_COOKIE	cookie;
	DAT_RETURN		status;
	DAT_EVENT		event;
	DAT_COUNT		nmore;
	DAT_LMR_TRIPLET		l_iov;
	DAT_RMR_TRIPLET		r_iov;
	volatile DAT_UINT64	*target = (DAT_UINT64*)buf[RCV_RDMA_BUF_INDEX];
	DAT_DTO_COMPLETION_EVENT_DATA *dto_event = 
		&event.event_data.dto_completion_event_data;
	DAT_IB_EXTENSION_EVENT_DATA *ext_event = 
		(DAT_IB_EXTENSION_EVENT_DATA *)&event.event_extension_data[0];

	printf("\nDoing FETCH and ADD\n");

        /* RMR info already swapped back to host order in connect_ep */
	r_iov = *buf[ RECV_BUF_INDEX ];
	
	l_iov.lmr_context     = lmr_atomic_context;
	l_iov.virtual_address = (DAT_UINT64)atomic_buf;
	l_iov.segment_length  = BUF_SIZE_ATOMIC;

	cookie.as_64 = 0x7777;
	if (server) {
		/* Wait for client to finish cmp_swap */
		while (*target != 0x98765)
			sleep(1);
		*target = 0x10;
		sleep(1);
		status = dat_ib_post_fetch_and_add(
						ep, 
						(DAT_UINT64)0x100,
						&l_iov,
						cookie, 
						&r_iov, 
						DAT_COMPLETION_DEFAULT_FLAG);
	} else {
		/* Wait for server, no swap so nothing to check */
		*target = 0x100;
		sleep(1);
		status = dat_ib_post_fetch_and_add(
						ep, 
						(DAT_UINT64)0x10, 
						&l_iov,
						cookie, 
						&r_iov, 
						DAT_COMPLETION_DEFAULT_FLAG);
	}
	_OK(status, "dat_ib_post_fetch_and_add");
	status = dat_evd_wait(dto_evd, DTO_TIMEOUT, 1, &event, &nmore);
	_OK(status, "dat_evd_wait for fetch and add");
	if (event.event_number != DAT_IB_DTO_EVENT) {
		printf("unexpected event after post_fetch_and_add: 0x%x\n",
			event.event_number); 
		exit(1);
	}

	_OK(dto_event->status, "event status for FETCH and ADD");
	if (ext_event->type != DAT_IB_FETCH_AND_ADD) {
		printf("unexpected event data of fetch and add : type=%d "
			"cookie=%d original%d\n",
			(int)ext_event->type,
			(int)dto_event->user_cookie.as_64,
			(int)*atomic_buf);
		exit(1);
	}

	if (server) {
	    printf("Client original data (on server) = "F64x", expected "
			"0x100\n", *atomic_buf);
	} else {
	    printf("Server original data (on client) = "F64x", expected "
			"0x10\n", *atomic_buf);
	}

	sleep(1);	

	if (server) {
		status = dat_ib_post_fetch_and_add(
						ep, 
						(DAT_UINT64)0x100, 
						&l_iov,
						cookie, 
						&r_iov, 
						DAT_COMPLETION_DEFAULT_FLAG);
	} else {
		status = dat_ib_post_fetch_and_add(
						ep, 
						(DAT_UINT64)0x10,
						&l_iov,
						cookie, 
						&r_iov, 
						 DAT_COMPLETION_DEFAULT_FLAG);
	}

	status = dat_evd_wait(dto_evd, DTO_TIMEOUT, 1, &event, &nmore);
	_OK(status, "dat_evd_wait for second fetch and add");
	if (event.event_number != DAT_IB_DTO_EVENT) {
		printf("unexpected event after second post_fetch_and_add: "
			"0x%x\n", event.event_number); 
		exit(1);
	}

	_OK(dto_event->status, "event status for second FETCH and ADD");
	if (ext_event->type != DAT_IB_FETCH_AND_ADD) {
		printf("unexpected event data of second fetch and add : "
			"type=%d cookie=%d original%p\n",
			(int)ext_event->type,
			(int)dto_event->user_cookie.as_64,
			atomic_buf);
		exit(1);
	}

	sleep(1); /* wait for other side to complete fetch_add */

	if (server) {
	    printf("Server got original data         = "F64x", expected "
			"0x200\n", *atomic_buf);
	    printf("Client final result (on server)  = "F64x", expected "
			"0x30\n", *target);

    	    if (*atomic_buf != 0x200 || *target != 0x30) {
	    	printf("ERROR: Server FETCH_ADD\n");
		exit(1);
	    }
	} else {
	    printf("Server side original data        = "F64x", expected "
			"0x20\n", *atomic_buf);
	    printf("Server final result (on client)  = "F64x", expected "
			"0x300\n", *target);

    	    if (*atomic_buf != 0x20 || *target != 0x300) {
	    	printf("ERROR: Server FETCH_ADD\n");
		exit(1);
	    }
	}
	printf("\n FETCH_ADD test - PASSED\n");
	return(0);
}

int
main(int argc, char **argv)
{
	char *hostname;
	int rc=0;

	if (argc > 2) {
		printf(usage);
		exit(1);
	}

	if ((argc == 1) || strcmp(argv[ 1 ], "-s") == 0)
	{
		server = 1;
	} else {
		server = 0;
		hostname = argv[ 1 ];
	}

#if defined(_WIN32) || defined(_WIN64)
	{
		WSADATA wsaData; 
		int i;

		i = WSAStartup ( MAKEWORD(2,2), &wsaData );
		if ( i != 0 ) {
			printf("%s WSAStartup(2.2) fail? (0x%x)\n",argv[0],i);
			fflush(stdout);
			exit(1);
		}
	}
#endif

	/*
	 * connect
	 */
	if (connect_ep(hostname)) {
		_WSACleanup();
		exit(1);
	}
	if (do_immediate()) {
		_WSACleanup();
		exit(1);
	}
	if (do_cmp_swap()) {
		_WSACleanup();
		exit(1);
	}
	if (do_fetch_add()) {
		_WSACleanup();
		exit(1);
	}
	rc = disconnect_ep();
	_WSACleanup();
	return rc;
}
